/* sevencolors.c
 *
 * Copyright (C) 2000 Stefano Bellotti
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>

#include "sevencolors.h"
#include "sevencolors_private.h"


/*
 * Public part
 */


/*
 * Creates a new allocated game table of the given size.
 * Rhombs colors are random.
 * The two corners always have different colors.
 */
GameTable *game_table_new (int width, int height)
{
  GameTable *table;
  int i;

  g_return_val_if_fail (width >= 2, NULL);
  g_return_val_if_fail (height >= 2, NULL);

  /* table allocation */
  table = (GameTable *) g_malloc (sizeof (GameTable) );

  /* table initializazion */
  table->height = height;
  table->width = width;

  /* Rhomb data allocation */
  table->rhomb = (Rhomb **) g_malloc (width * sizeof (Rhomb *) );
  for(i=0; i<width; i++)
    table->rhomb[i] = (Rhomb *) g_malloc (height * sizeof (Rhomb) );

  /* Rhomb data initialization */
  game_table_randomize (table);

  return table;
}

/*
 * Relases all memory used by the game table.
 */
void game_table_destroy (GameTable *table)
{
  int i;

  g_return_if_fail (table != NULL);

  /* free Rhomb data */
  for (i = 0; i < table->width; i++) g_free (table->rhomb[i]);
  g_free (table->rhomb);

  /* free table */
  g_free (table);
}

/*
 * Randomizes the colors of a game table.
 * Rhombs colors are random.
 * The two corners always have different colors.
 */
void game_table_randomize (GameTable *table)
{
  int x,y;

  g_return_if_fail (table != NULL);

  /* random colors */
  for (y = 0; y < table->height; y++)
    for (x = 0; x < table->width; x++)
      table->rhomb[x][y].col = 7.0 * rand () / (RAND_MAX + 1.0);

 /* check that players have different colors */
 if(table->rhomb[0][table->height - 1].col == table->rhomb[table->width - 1][0].col)
   table->rhomb[table->width - 1][0].col = (table->rhomb[table->width - 1][0].col + 1) % 7;
}

/*
 * Gets the color of the corner rhomb.
 */
int game_table_get_color (GameTable *table, GameTableCorner corner)
{
  int x,y;

  g_return_val_if_fail (table != NULL, 0);

  game_table_corner_to_xy (table, corner, &x, &y);

  return table->rhomb[x][y].col;
}

/*
 * Gets the number of rhombs owned by the player in that corner.
 */
int game_table_get_number_of_rhombs (GameTable *table, GameTableCorner corner)
{
  int x,y;

  g_return_val_if_fail (table != NULL, 0);

  game_table_corner_to_xy (table, corner, &x, &y);
  game_table_clear (table);

  return game_table_expand (table, x, y, table->rhomb[x][y].col);
}

/*
 * The color expands from the given corner.
 * Returns the total number of rhombs that get colored.
 */
int game_table_play_color (GameTable *table, int color, GameTableCorner corner)
{
  int k,x,y;

  g_return_val_if_fail (table != NULL, 0);
  g_return_val_if_fail (color >= 0, 0);
  g_return_val_if_fail (color < 7, 0);
  g_return_val_if_fail (color != game_table_get_color (table, corner), 0);
  g_return_val_if_fail (color != game_table_get_color (table, (corner +1) % 2), 0);

  /* get player position */
  game_table_corner_to_xy (table, corner, &x, &y);
  /* clear tags */
  game_table_clear (table);
  /* paint and tag rhombs */
  k = game_table_paint (table, x, y, table->rhomb[x][y].col, color);
  /* get opponent position */
  game_table_corner_to_xy (table, (corner + 1) % 2 , &x, &y);
  /* tag all reachable rhombs */
  game_table_mark_reachable (table, x, y);
  /* paint remainings rhobmbs */
  k += game_table_paint_not_marked (table, color);

  return k;
}

/*
 * The ai chooses the best color.
 * Returns the choosen color.
 * TODO: implement a more intelligent ai.
 */
int game_table_computer_choose_color (GameTable *table, GameTableCorner corner)
{
  int x, y, col, col_max, score, score_max;

  g_return_val_if_fail (table != NULL, 0);

  game_table_corner_to_xy (table, corner, &x, &y);

  score_max = 0;
  col_max = 0;

  for (col = 0; col < 7; col++)
  {
    if (col != game_table_get_color (table, CORNER_LOW_LEFT) &&
        col != game_table_get_color (table, CORNER_TOP_RIGHT) )
    {
      score = game_table_what_if (table, corner, col);
      if (score >= score_max)
      {
       score_max = score;
       col_max = col;
      }
    }
  }

  return col_max;
}

/*
 * The ai chooses and plays the best color.
 * Returns the total number of rhombs that get colored and the choosen color.
 * TODO: implement a more intelligent ai.
 */
int game_table_computer_play (GameTable *table, GameTableCorner corner, int *color)
{
  g_return_val_if_fail (table != NULL, 0);

  *color = game_table_computer_choose_color (table, corner);
  return game_table_play_color (table, *color, corner);
}

/*
 * Returns the total number of rhombs that would get colored if the player
 * in the given corner choses the given color.
 * The game table is not modified.
 */
int game_table_what_if (GameTable *table, GameTableCorner corner, int color)
{
  int k, x, y;

  g_return_val_if_fail (table != NULL, 0);
  g_return_val_if_fail (color >= 0, 0);
  g_return_val_if_fail (color < 7, 0);

  /* get player position */
  game_table_corner_to_xy (table, corner, &x, &y);
  /* clear tags */
  game_table_clear (table);
  /* tag rhombs */
  k = game_table_paint_fake (table, x, y, table->rhomb[x][y].col, color);
  /* get opponent position */
  game_table_corner_to_xy (table, (corner + 1) % 2 , &x, &y);
  /* tag all reachable rhombs */
  game_table_mark_reachable (table, x, y);
  /* tag remainings rhobmbs */
  k += game_table_paint_not_marked_fake (table);

  return k;
}


/*
 * Private part
 */


/*
 * Return the coordinates (x,y) of a corner.
 */
void game_table_corner_to_xy (GameTable *table, GameTableCorner corner, int *x, int *y)
{
  g_return_if_fail (table != NULL);

  switch (corner)
  {
    case CORNER_LOW_LEFT:
      *x = 0;
      *y = table->height - 1;
      break;
    case CORNER_TOP_RIGHT:
      *x = table->width - 1;
      *y = 0;
      break;
    default:
      g_assert_not_reached ();
  }
}

/*
 * Reset all the flags.
 */
void game_table_clear (GameTable *table)
{
  int i, j;

  g_return_if_fail (table != NULL);

  for(i = 0; i < table->width ; i++)
    for(j = 0; j < table->height ; j++)
      table->rhomb[i][j].flag = 0;
}

/*
 * Recursvive function that, starting from (x,y), change the color of all
 * reachable rhombs of color old_col to new_color.
 * For all neighbour rhombs of color old_color the recursive function
 * game_table_expand is called.
 * Returns the number of all acquired rhombs.
 */
int game_table_paint (GameTable *table, int x, int y, int old_col, int new_col)
{
  int k = 0;

  g_return_val_if_fail (table != NULL, 0);
  g_return_val_if_fail (old_col >= 0, 0);
  g_return_val_if_fail (old_col < 7, 0);
  g_return_val_if_fail (new_col >= 0, 0);
  g_return_val_if_fail (new_col < 7, 0);

  /* check end of recursion */
  if (x < table->width && x >= 0 && y < table->height && y >= 0)
  {
    /* go on painting */
    if (table->rhomb[x][y].col == old_col)
    {
      /* paint rhomb*/
      table->rhomb[x][y].col = new_col;
      /* mark rhomb */
      table->rhomb[x][y].flag = 1;

      /* even line */
      if (y%2 == 0)
        k=game_table_paint (table, x, y - 1, old_col, new_col) +
          game_table_paint (table, x, y + 1, old_col, new_col) +
          game_table_paint (table, x + 1, y - 1, old_col, new_col) +
          game_table_paint (table, x + 1, y + 1, old_col, new_col) + 1;
      /* odd line */
      else
        k=game_table_paint (table, x, y - 1, old_col, new_col) +
          game_table_paint (table, x, y + 1, old_col, new_col) +
          game_table_paint (table, x - 1, y - 1, old_col, new_col) +
          game_table_paint (table, x - 1, y + 1, old_col, new_col) + 1;
    }
    /* get new rhombs*/
    else
    {
      /* check if it's a rhomb to count */
      if (table->rhomb[x][y].col == new_col && table->rhomb[x][y].flag == 0)
      {
        /* mark rhomb */
        table->rhomb[x][y].flag = 1;
        /* even line */
        if (y%2 == 0)
          k=game_table_expand (table, x, y - 1, new_col) +
            game_table_expand (table, x, y + 1, new_col) +
            game_table_expand (table, x + 1, y - 1, new_col) +
            game_table_expand (table, x + 1, y + 1, new_col) + 1;
        /* odd line */
        else
          k=game_table_expand (table, x, y - 1, new_col) +
            game_table_expand (table, x, y + 1, new_col) +
            game_table_expand (table, x - 1, y - 1, new_col) +
            game_table_expand (table, x - 1, y + 1, new_col) + 1;

      }
    }
  }

  return k;
}

/*
 * Like game_table_paint, but this function doesn't change the color, it sets
 * only the flag.
 */
int game_table_paint_fake (GameTable *table, int x, int y, int old_col, int new_col)
{
  int k = 0;

  g_return_val_if_fail (table != NULL, 0);
  g_return_val_if_fail (old_col >= 0, 0);
  g_return_val_if_fail (old_col < 7, 0);
  g_return_val_if_fail (new_col >= 0, 0);
  g_return_val_if_fail (new_col < 7, 0);

  /* check end of recursion */
  if (x < table->width && x >= 0 && y < table->height && y >= 0)
  {
    /* go on painting */
    if (table->rhomb[x][y].col == old_col && table->rhomb[x][y].flag == 0)
    {
      /* mark rhomb */
      table->rhomb[x][y].flag = 1;

      /* even line */
      if (y%2 == 0)
        k=game_table_paint_fake (table, x, y - 1, old_col, new_col) +
          game_table_paint_fake (table, x, y + 1, old_col, new_col) +
          game_table_paint_fake (table, x + 1, y - 1, old_col, new_col) +
          game_table_paint_fake (table, x + 1, y + 1, old_col, new_col) + 1;
      /* odd line */
      else
        k=game_table_paint_fake (table, x, y - 1, old_col, new_col) +
          game_table_paint_fake (table, x, y + 1, old_col, new_col) +
          game_table_paint_fake (table, x - 1, y - 1, old_col, new_col) +
          game_table_paint_fake (table, x - 1, y + 1, old_col, new_col) + 1;
    }
    /* get new rhombs*/
    else
    {
      /* check if it's a rhomb to count */
      if (table->rhomb[x][y].col == new_col && table->rhomb[x][y].flag == 0)
      {
        /* mark rhomb */
        table->rhomb[x][y].flag = 1;
        /* even line */
        if (y%2 == 0)
          k=game_table_expand (table, x, y - 1, new_col) +
            game_table_expand (table, x, y + 1, new_col) +
            game_table_expand (table, x + 1, y - 1, new_col) +
            game_table_expand (table, x + 1, y + 1, new_col) + 1;
        /* odd line */
        else
          k=game_table_expand (table, x, y - 1, new_col) +
            game_table_expand (table, x, y + 1, new_col) +
            game_table_expand (table, x - 1, y - 1, new_col) +
            game_table_expand (table, x - 1, y + 1, new_col) + 1;

      }
    }
  }

  return k;
}

/*
 * Recursvive function that set the flag of all rhombs of color col reachable
 * from (x,y).
 * Returns the number of marked rhombs.
 */
int game_table_expand (GameTable *table, int x, int y, int col)
{
  int k = 0;

  g_return_val_if_fail (table != NULL, 0);
  g_return_val_if_fail (col >= 0, 0);
  g_return_val_if_fail (col < 7, 0);

  /* check end of recursion */
  if (x < table->width && x >= 0 && y < table->height && y >= 0)
  {
    /* go on expanding */
    if (table->rhomb[x][y].col == col && table->rhomb[x][y].flag == 0)
    {
      /* mark rhomb */
      table->rhomb[x][y].flag = 1;
      /* even line */
      if (y%2 == 0)
        k=game_table_expand (table, x, y - 1, col) +
          game_table_expand (table, x, y + 1, col) +
          game_table_expand (table, x + 1, y - 1, col) +
          game_table_expand (table, x + 1, y + 1, col) + 1;
      /* odd line */
      else
        k=game_table_expand (table, x, y - 1, col) +
          game_table_expand (table, x, y + 1, col) +
          game_table_expand (table, x - 1, y - 1, col) +
          game_table_expand (table, x - 1, y + 1, col) + 1;
    }
  }

  return k;
}

/*
 * Recursvive function that set the flag of all rhombs reachable from (x,y)
 * without passing on already market rhombs.
 */
void game_table_mark_reachable (GameTable *table, int x, int y)
{
  g_return_if_fail (table != NULL);

  /* check end of recursion */
  if (x < table->width && x >= 0 && y < table->height && y >= 0)
  {
    /* go on expanding */
    if (table->rhomb[x][y].flag == 0)
    {
      /* mark rhomb */
      table->rhomb[x][y].flag = 1;
      /* even line */
      if (y%2 == 0)
      {
        game_table_mark_reachable (table, x, y - 1);
        game_table_mark_reachable (table, x, y + 1);
        game_table_mark_reachable (table, x + 1, y - 1);
        game_table_mark_reachable (table, x + 1, y + 1);
      }
      /* odd line */
      else
      {
        game_table_mark_reachable (table, x, y - 1);
        game_table_mark_reachable (table, x, y + 1);
        game_table_mark_reachable (table, x - 1, y - 1);
        game_table_mark_reachable (table, x - 1, y + 1);
      }
    }
  }
}

/*
 * Change the color of all not marked rhombs.
 * Returns the number of painted rhombs.
 */
int game_table_paint_not_marked (GameTable *table, int col)
{
  int k = 0, i ,j;

  g_return_val_if_fail (table != NULL, 0);
  g_return_val_if_fail (col >= 0, 0);
  g_return_val_if_fail (col < 7, 0);

  for(i = 0; i < table->width ; i++)
    for(j = 0; j < table->height ; j++)
      if (table->rhomb[i][j].flag == 0)
      {
        k++;
        table->rhomb[i][j].col = col;
        table->rhomb[i][j].flag = 1 ;
      }

  return k;
}

/*
 * Like game_table_paint_not_marked, but this function doesn't change the
 * color, it sets only the flag.
 */
int game_table_paint_not_marked_fake (GameTable *table)
{
  int k = 0, i ,j;

  g_return_val_if_fail (table != NULL, 0);

  for(i = 0; i < table->width ; i++)
    for(j = 0; j < table->height ; j++)
      if (table->rhomb[i][j].flag == 0)
      {
        k++;
        table->rhomb[i][j].flag = 1 ;
      }

  return k;
}
